(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Check_Conditional_CRP
imports 
  Check_CRP
begin

datatype ('f, 'l, 'v) conditional_cr_proof =
  Unravel_CR 
    "((('f, 'l) lab, 'v) crule \<times> (('f, 'l) lab, 'v) rules) list"
    "('f, 'l, 'v) cr_proof" |
  Almost_Orthogonal_CR |
  Almost_Orthogonal_Modulo_Infeasibility_CR
    "((('f, 'l) lab, 'v) rules \<times> (('f, 'l) lab, 'v) infeasibility_proof) list"

primrec
  check_conditional_cr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f, label_type) lab, string) crules \<Rightarrow> 
     ('f, label_type, string) conditional_cr_proof \<Rightarrow>
     shows check" 
where
  "check_conditional_cr_proof a i I J ctrs (Unravel_CR u_info prf) = debug (i []) ''Unravel'' (do {
    r \<leftarrow> check_sp_unraveling u_info ctrs
      <+? (\<lambda> s. i \<circ> shows '': error in unraveling'' \<circ> shows_nl \<circ> indent s);
    check_cr_proof a (i \<circ> shows_string ''.1'') I J r prf
      <+? (\<lambda>s. i \<circ> shows '': error below confluence proof'' \<circ> shows_nl \<circ> indent s)
  })" |
  "check_conditional_cr_proof a i I J ctrs Almost_Orthogonal_CR =
    debug (i []) ''Almost-Orthogonal'' (check_level_confluence ctrs)" |
  "check_conditional_cr_proof a i I J ctrs (Almost_Orthogonal_Modulo_Infeasibility_CR cps) =
    debug (i []) ''Almost-Orthogonal modulo Infeasibility'' (
      check_level_confluence_modulo_infeasibility cps ctrs)"

primrec
  conditional_cr_assms :: "bool \<Rightarrow> ('f, 'l, 'v) conditional_cr_proof \<Rightarrow> (('f, 'l, 'v) assm) list"
where
  "conditional_cr_assms a (Unravel_CR _ p) = cr_assms a p" |
  "conditional_cr_assms a Almost_Orthogonal_CR = []" |
  "conditional_cr_assms a (Almost_Orthogonal_Modulo_Infeasibility_CR _) = []"

lemma check_conditional_cr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (conditional_cr_assms a prf). holds p"
    and ok: "isOK (check_conditional_cr_proof a i I J ctrs prf)"
  shows "CR (cstep (set ctrs))"
  using fin ok
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (Unravel_CR u_info prof)
    note check = Unravel_CR(1)[simplified]
    from check obtain R where R: "check_sp_unraveling u_info ctrs = return R" by auto
    from check[unfolded R]
    have ok: "isOK (check_cr_proof a (i +@+ shows_string ''.1'') I J R prof)" by simp
    from Unravel_CR(2) have a: "\<forall>a\<in>set (cr_assms a prof). holds a" by auto
    from check_cr_proof_with_assms_sound[OF I J a ok] have CR: "CR (rstep (set R))" by simp
    show ?case
      by (rule check_sp_unraveling_CR[OF R infinite_lab CR])
  next
    case Almost_Orthogonal_CR
    then show ?case by (intro level_confluent_imp_CR; simp)
  next
    case (Almost_Orthogonal_Modulo_Infeasibility_CR cps)
    then show ?case by (intro level_confluent_imp_CR; simp)
  qed
qed

lemma conditional_cr_assms_False [simp]:
  "conditional_cr_assms False prf = []"
  by (induct "prf") simp_all

lemma check_conditional_cr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and ok: "isOK (check_conditional_cr_proof False i I J ctrs prf)"
  shows "CR (cstep (set ctrs))"
  by (rule check_conditional_cr_proof_with_assms_sound[OF I J _ ok], simp)

end
